/* Auteur - La Massiot - 2005 */

#include "couche_logique.h"
#include <string.h>
#include <strings.h>
#include <assert.h>

int    nb_files_glob;
int    nb_data_blocks_glob;

char * tab_of_allocated_file_blocks_glob;
char * tab_of_allocated_data_blocks_glob;

/*--------------------------------------*/
/*          u16_to_big_endian           */
/*--------------------------------------*/
u16 u16_to_big_endian(u16 a)
{
  u8 b [2];

    b[0] = (a/256) % 256;
    b[1] = a % 256;

    return *(u16 *) b;
}
/*--------------------------------------*/
/*          u32_to_big_endian           */
/*--------------------------------------*/
u32 u32_to_big_endian(u32 a)
{
  u8 b [4];

    b[0] = (((a/256)/256)/256) % 256;
    b[1] = ((a/256)/256) % 256;
    b[2] = (a/256) % 256;
    b[3] = a % 256;

    return *(u32 *) b;
}
/*--------------------------------------*/
/*          big_endian_to_u16           */
/*--------------------------------------*/
u16 big_endian_to_u16(u16 a)
{
  u8 b[2];
  *(u16 *) b = a;

  return 256*b[0] + b[1];
}
/*--------------------------------------*/
/*          big_endian_to_u32           */
/*--------------------------------------*/
u32 big_endian_to_u32(u32 a)
{
  u8 b[4];
  
  *(u32 *) b = a;
 
  return 256*256*256*b[0] + 256*256*b[1] + 256*b[2] + b[3];
}
/*----------------------------------------------------*/
/*               init_couche_logique                  */
/*----------------------------------------------------*/
/*
On suppose que le disque est format (l'appel  la fontion
"format" de la couche applications a eu lieu).

- Valuation des constantes "nb_files_glob" et "nb_data_blocks_glob"
- Initialisation du cache
- Initialisation des tableaux "tab_of_allocated_file_blocks_glob"
  et "tab_of_allocated_data_blocks_glob".
*/
error_t init_couche_logique()
{
error_t e = 0;  
struct s_partition_de_description block_de_descr;
  
  if((e = start_disk()) != 0)
    return e;
  
  /* LECTURE BLOC DE DESCRITPION */
  if((e = read_pblock(0, (pblock_t) &block_de_descr)) != 0)
    return e;
  if((e = read_pblock(1, (pblock_t) (SIZEOF_PBLOCK + (char *) &block_de_descr))) != 0)
    return e;

  /* VALUATION DES VARIABLES "nb_files_glob" et
     "nb_data_blocks_glob" */
  nb_files_glob       = big_endian_to_u32(block_de_descr.pd_n);
  
  nb_data_blocks_glob = big_endian_to_u32(block_de_descr.pd_nb_of_free_blocks);

  /* INITIALISATION DU CACHE */
  if((e = init_cache(nb_files_glob + nb_data_blocks_glob + 1)) != 0)
    return e; 

  /* ALLOCATION DES TABLEAUX
     "tab_of_allocated_file_blocks_glob" et
     "tab_of_allocated_data_blocks_glob" */
  if((e = init_tabs_of_free_blocks()) != 0)
  {
    shutdown_cache();
    return e; 
  }
      
  /* INITIALISATION DES TABLEAUX 
     "tab_of_allocated_file_blocks_glob" et
     "tab_of_allocated_data_blocks_glob" */
  if((e = update_tabs_of_free_blocks() ) != 0)
    return e;

  return 0;
}
/*----------------------------------------------------------*/
/*   OK               shutdown_couche_logique                 */
/*----------------------------------------------------------*/
/*
- Dsalloue les deux tableaux
- Appelle shutdown_cache (cache.c)
*/
error_t shutdown_couche_logique()
{
  free(tab_of_allocated_file_blocks_glob);
  free(tab_of_allocated_data_blocks_glob);
  return shutdown_cache();
}
/*----------------------------------------------------------*/
/*                  init_tabs_of_free_blocks                */
/*----------------------------------------------------------*/
/*
Allocation de "tab_of_allocated_file_blocks_glob" :
tableau de "nb_files_glob" char.

Allocation de "tab_of_allocated_data_blocks_glob" :
tableau de "nb_data_blocks_glob" char. 
*/
error_t init_tabs_of_free_blocks()
{
  tab_of_allocated_file_blocks_glob = (char *) malloc(sizeof(char)*nb_files_glob);
 
  if(tab_of_allocated_file_blocks_glob == NULL)
    return ERROR_MALLOC;
  
  tab_of_allocated_data_blocks_glob = (char *) malloc(sizeof(char)*nb_data_blocks_glob);
  
  if(tab_of_allocated_data_blocks_glob == NULL)
    return ERROR_MALLOC;

  return 0;
}
/*----------------------------------------------------------*/
/*                  update_tabs_of_free_blocks              */
/*----------------------------------------------------------*/
/*
Cette fonction permet de mettre  jour les tableaux
- tab_of_allocated_file_blocks_glob
- tab_of_allocated_data_blocks_glob
en regardant sur le disque si les blocs sont libres ou pas.

 
- On lit les blocs de fichier un par un.
- On regarde s'ils sont libres.
- Si un bloc de fichier n'est pas libre, on marque un "1" pour 
 dire qu'il est occup dans "tab_of_allocated_file_blocks_glob"
 et
 d'aprs sa taille, on en dduit le nombre de blocs de donnes
 qu'il a en propre "nb_data_blocks"
 on remplit avec des "1" "tab_of_allocated_data_blocks_glob"
 dans les cases correspondant aux numros des blocs propres
  ce fichier.

Remarque :
On lit les blocs de fichier pour rcuprer :
- bloc.pf_state
- bloc.pf_size
- bloc.pf_data
*/
error_t update_tabs_of_free_blocks()
{
struct s_fs_inode bloc;
              int i;
              int j;
          error_t e;
              int nb_data_blocks;
  
  /* On met les deux tableaux  zro */
  bzero(tab_of_allocated_file_blocks_glob, nb_files_glob * sizeof(char));
  bzero(tab_of_allocated_data_blocks_glob, nb_data_blocks_glob * sizeof(char));

  for(i=0 ; i<nb_files_glob ; i++)
  {
    /* LECTURE */
    if((e = read_sblock(i+1, (sblock_t) &bloc)) != 0)
      return e;

    /* MARQUAGE */
    if(big_endian_to_u16(bloc.pf_state) == STATE_OCCUPIED)
    {
      tab_of_allocated_file_blocks_glob[i] = 1;

      /* Nombre de blocs de donnes */
      nb_data_blocks = nb_blocks_given_the_size_of_a_file(big_endian_to_u16(bloc.pf_size));

      assert(0 <= nb_data_blocks && nb_data_blocks <= NB_BLOCKS_MAX_PER_FILE);

      /* Blocs de donnes associs a ce fichier */
      for(j=0 ; j<nb_data_blocks ; j++)
      {
        int k;
        k = big_endian_to_u16(bloc.pf_data[j]);
        assert(0 <= k && k < nb_data_blocks_glob);
        
        /* MARQUAGE */
        tab_of_allocated_data_blocks_glob[k - nb_files_glob - 1] = 1; 
      }
    }
  }
  return 0;
}
/*----------------------------------------------------------*/
/*            nb_blocks_given_the_size_of_a_file            */
/*----------------------------------------------------------*/
int nb_blocks_given_the_size_of_a_file(int size)
{
  if((size % SIZE_OF_SBLOCK) == 0)
    return size/SIZE_OF_SBLOCK;

  else
    return size/SIZE_OF_SBLOCK + 1;
}
/*----------------------------------------------------------*/
/*                  find_new_file_block_number              */
/*----------------------------------------------------------*/
/*
Trouve le premier numro de bloc de fichiers libre dans 
"tab_of_allocated_file_blocks_glob".

On retourne dans "p_result" l'indice (plus 1) de la premire 
case de "tab_of_allocated_file_blocks_glob" qui contient un "0".
*/
error_t find_new_file_block_number(int * p_result)
{
int i;

  i = 0;
  while(i < nb_files_glob)
  {
    if(tab_of_allocated_file_blocks_glob[i] == 0)
    {
      *p_result = 1 + i;
      return 0;
    }
    i++;
  }

  return ERROR_NO_MORE_AVAILABLE_FILES_BLOCKS;
}
/*----------------------------------------------------------*/
/*                  find_new_data_block_number              */
/*----------------------------------------------------------*/
/* Trouve le premier bloc de donnes disponible pour un 
   fichier. */

error_t find_new_data_block_number(struct s_fs_inode * p_inode, int * p_result)
{
int nb_blocks;
int nb_of_the_last_block;
int nb_of_the_last_block_relative;
int i;

  /* COMBIEN DE BLOCS DE DONNEES ? */
  /* On regarde combien le fichier a de blocs de donnes en
     propre. */
  nb_blocks = nb_blocks_given_the_size_of_a_file(
              big_endian_to_u16(p_inode->pf_size));

  /*--- AUCUN BLOC DE DONNEES ---*/
  /* On recherche le premier bloc de donnes libre ( 0)
     dans "tab_of_allocated_data_blocks_glob" */
  if(nb_blocks == 0)
  {
    for(i=0 ; i<nb_data_blocks_glob ; i++)
    {
      if(tab_of_allocated_data_blocks_glob[i] == 0)
      {
        *p_result = 1 + nb_files_glob + i; 
        return 0;
      }
    }
  }

  /*--- UN OU PLUSIEURS BLOCS DE DONNEES ---*/
  /* On recherche dans "tab_of_allocated_data_blocks_glob" le
     numro de bloc de donnes le plus proche du dernier bloc
     de donnes du fichier */
  else
  {
    nb_of_the_last_block = big_endian_to_u16(p_inode->pf_data[nb_blocks - 1]);
    nb_of_the_last_block_relative = nb_of_the_last_block - 1 - nb_files_glob;
    
    for(i=nb_of_the_last_block_relative ; i<nb_data_blocks_glob ; i++)
    {
      if(tab_of_allocated_data_blocks_glob[i] == 0)
      {
        *p_result = 1 + nb_files_glob + i; 
        return 0;
      }
    }
    
    /* On n'a pas trouv de bloc de donnes libre dans la
       partie du tableau comprise entre les indices
       "nb_of_the_last_block_relative" et "nb_data_blocks_glob"
      
       ICI ON REPREND LA RECHERCHE AU DEBUT DU TABLEAU
       On s'arrte avant la fin du tableau 
    */
    for(i=0 ; i<nb_of_the_last_block_relative ; i++)
    {
      if(tab_of_allocated_data_blocks_glob[i] == 0)
      {
        *p_result = 1 + nb_files_glob + i; 
        return 0;
      }
    }
  }
    
  /* PAS DE BLOC DE DONNEES LIBRE */
  return ERROR_NO_MORE_AVAILABLE_DATA_BLOCKS;
}
/*-----------------------------------------------------*/
/*               alloc_new_file_block                  */
/*-----------------------------------------------------*/
/*
Dclare occup dans "tab_of_allocated_file_blocks_glob"
le bloc que l'on alloue.

Le premier bloc de fichier a pour numro 1.
*/
void alloc_new_file_block(int nro_physique)
{
  /* Le premier numro physique de bloc de fichier est 1. */
  assert(0 < nro_physique && nro_physique <= nb_files_glob);
  assert(tab_of_allocated_file_blocks_glob[nro_physique - 1] == 0);

  tab_of_allocated_file_blocks_glob[nro_physique - 1] = 1;
}
/*-----------------------------------------------------*/
/*               alloc_new_data_block                  */
/*-----------------------------------------------------*/
/*
Dclare occup dans "tab_of_allocated_data_blocks_glob"
le bloc que l'on alloue.

Le premier bloc de donnes a pour numro nb_files_glob + 1.
"tab_of_allocated_data_blocks_glob" ne comporte que 
"nb_data_blocks_glob" cases.

*/
void alloc_new_data_block(int nro_physique)
{
  assert(nb_files_glob + 1 <= nro_physique && nro_physique <= nb_files_glob + nb_data_blocks_glob + 1);
  assert(tab_of_allocated_data_blocks_glob[nro_physique - nb_files_glob - 1] == 0);
  
  tab_of_allocated_data_blocks_glob[nro_physique - nb_files_glob - 1] = 1;
}
/*-----------------------------------------------------*/
/*                  desalloc_data_block                */
/*-----------------------------------------------------*/
/*
Dclare libre dans "tab_of_allocated_data_blocks_glob"
le bloc que l'on dsalloue.
*/
void desalloc_data_block(int nro_physique)
{
  assert(nb_files_glob + 1 <= nro_physique && nro_physique <= nb_files_glob + nb_data_blocks_glob + 1);
  assert(tab_of_allocated_data_blocks_glob[nro_physique - nb_files_glob - 1] == 1);

  tab_of_allocated_data_blocks_glob[nro_physique - nb_files_glob - 1] = 0;
}
/*-----------------------------------------------------*/
/*              update_description_block               */
/*-----------------------------------------------------*/
/*
- Compte dans "tab_of_allocated_file_blocks_glob" le nombre 
  de blocs de fichiers libres.
- Compte dans "tab_of_allocated_data_blocks_glob" le nombre
  de blocs de donnes libres.
- Mmorise le numro du premier bloc de donnes libre.

Met  jour la partition de description :
champs "pd_nb_of_free_files"
       "pd_nb_of_free_blocks"
       "pd_nro_first_free_block"
*/

error_t update_description_block()
{
int nb_of_free_file_blocks;
int nb_of_free_data_blocks;
int nro_of_first_free_data_block;
int i;
error_t e;
struct s_partition_de_description description_block;
  

  nb_of_free_file_blocks = 0;
  nb_of_free_data_blocks = 0;
  nro_of_first_free_data_block = -1;
  
  /* LECTURE DU BLOC DE DESCRIPTION */
  if((e = read_sblock(0, (sblock_t) &description_block)) != 0)
    return e;

  /* NOMBRE DE BLOCS DE FICHIERS LIBRES */
  for(i=0 ; i<nb_files_glob ; i++)
  {
    if(tab_of_allocated_file_blocks_glob[i] == 0)
      nb_of_free_file_blocks++;   
  }
 
  /* Mise  jour du champ "pd_nb_of_free_files" (bloc de
     description) */
  description_block.pd_nb_of_free_files = u32_to_big_endian(nb_of_free_file_blocks);

  /* NOMBRE DE BLOCS DE DONNEES LIBRES */
  /* NUMERO DU PREMIER BLOC DE DONNEES LIBRE */
  for(i=0 ; i<nb_data_blocks_glob ; i++)
  {
    if(tab_of_allocated_data_blocks_glob[i] == 0)
    {
      if(nro_of_first_free_data_block == -1)
        nro_of_first_free_data_block = i + nb_files_glob + 1;
  
      nb_of_free_data_blocks++;
    }
  }
  
  /* Mise  jour des champs "pd_nb_of_free_blocks" et "pd_nro_first_free_block" */
  description_block.pd_nb_of_free_blocks    = u32_to_big_endian(nb_of_free_data_blocks);
  description_block.pd_nro_first_free_block = u32_to_big_endian(nro_of_first_free_data_block);

  return 0;
}
/*-------------------------------------*/
/*             create_file             */
/*-------------------------------------*/
/*
Crer un nouveau fichier, c'est simplement "rserver"
un nouveau bloc parmi les blocs de fichiers du disque
pour ce nouveau fichier.

Pour se faire, on appelle la fonction :
find_new_file_block_number
qui permet d'obtenir le premier bloc de fichier libre.

De plus,  sa cration, on spcifie :
- si le fichier est de type rgulier ou rpertoire,
- s'il peut_tre crit et/ou lu.
*/

error_t create_file(int type, int permissions, int * p_nb_physical_block_file)
{
          error_t  e;
struct s_fs_inode  new_file_block;

  bzero(&new_file_block, sizeof(struct s_fs_inode));
 
  /* TROUVE UN BLOC DE FICHIER LIBRE */
  /* Rcupration du numro de ce bloc dans 
     "p_nb_physical_block_file" */
  if((e = find_new_file_block_number(p_nb_physical_block_file)) != 0)
    return e;

  new_file_block.pf_protect = permissions;
  new_file_block.pf_type    = type;
  new_file_block.pf_size    = 0;
  new_file_block.pf_state   = u16_to_big_endian(STATE_OCCUPIED);
  new_file_block.pf_links   = 0;

  /* ECRITURE DANS LE CACHE DU BLOC DE FICHIER */
  if((e = write_sblock(*p_nb_physical_block_file, (sblock_t) &new_file_block )) != 0) 
    return e;

  /* ALLOCATION DU NOUVEAU FICHIER DANS 
     "tab_of_allocated_file_blocks_glob" */
  alloc_new_file_block(*p_nb_physical_block_file);
  
  return 0; 
}
/*------------------------------------------------------*/
/*         open_file_from_physical_block_number         */
/*------------------------------------------------------*/
/*
Ouvrir un fichier  partir du numro de bloc physique
qui lui correspond, c'est retourner une structure 
"struct s_file_descr" value comme il faut.
*/
error_t open_file_from_physical_block_number(int physical_block_number,
                                             int b_reading_authorized, 
                                             int b_writing_authorized, 
                                             int b_writing_at_the_end,
                                             fd * p_fd)
{
          error_t  e;
struct s_fs_inode  block;

  /* VERIFICATION */
  if(tab_of_allocated_file_blocks_glob[physical_block_number - 1] == 0)
   return ERROR_FILE_NOT_ALREADY_CREATED; 

  /* LECTURE DANS LE CACHE DU BLOC DE FICHIER */
  if((e = read_sblock(physical_block_number, (sblock_t) &block)) != 0)
    return e;
  
  /* ALLOCATION D'UN DESCRIPTEUR POUR CE FICHIER */
  *p_fd = (struct s_file_descr *) malloc(sizeof(struct s_file_descr));
 
  if(*p_fd == NULL)
    return ERROR_MALLOC;
        
  /*----- VALUATION ------*/

  (*p_fd)->nb_physical_block = physical_block_number;

  /* LECTURE AUTORISEE OU PAS - AVORTER ? */
  if((block.pf_protect & PERMIT_READ) != 0) /* On a le droit de lire. */
    (*p_fd)->b_reading_authorized  = b_reading_authorized;

  else 
  {
    /* On essaye d'ouvrir en lecture mais on n'a pas le droit :
       avorter */
    if(b_reading_authorized != 0) 
    {
      free(*p_fd);
      return ERROR_READING_NOT_AUTHORISED;
    } 
    
    /* On essaye pas d'ouvrir en lecture */
    else
      (*p_fd)->b_reading_authorized = 0;
  }
 
  /* ECRITURE AUTORISEE OU PAS - AVORTER ? */
  if((block.pf_protect & PERMIT_WRITE) != 0) /* On a le droit d'crire. */
    (*p_fd)->b_writing_authorized  = b_writing_authorized;

  else 
  {
    /* On essaye d'ouvrir en criture mais on n'a pas le droit :
       avorter */
    if(b_writing_authorized != 0)
    {
      free(*p_fd);
      return ERROR_WRITING_NOT_AUTHORISED;
    } 
    
    /* On essaye pas d'ouvrir en criture */
    else
      (*p_fd)->b_writing_authorized = 0;
  }
  
  (*p_fd)->size = big_endian_to_u16(block.pf_size);

  if(b_writing_at_the_end == 1)
    (*p_fd)->current_position = big_endian_to_u16(block.pf_size);

  else
    (*p_fd)->current_position = 0;

  return 0;
}
/*-------------------------------------*/
/*              read_file              */
/*-------------------------------------*/
/*
Lit "nb_bytes_to_read" dans le fichier de descripteur "file_descr"
 partir de la position courante dans le fichier ouvert.
Si c'est trop, on lit de la position courante  la fin du fichier.


Cette fonction prend pour arguments :
- un descripteur de fichier (d),
- un "buffer" dans lequel seront stocks les octets lus,
- le nombre d'octets  lire : "nb_bytes_to_read".

Ces octets seront lus  partir de la position courante
dans le fichier ouvert, 4me champ du descripteur.

Si "pc" est la position courante,
- alors nro_curent_blk = (pc / SIZE_OF_SBLOCK) est le 
numro du bloc courant
- et nro_current_byte_in_blk = (pc % SIZE_OF_SBLOCK)
est l'octet courant.

Le nombre d'octets restants  lire dans le bloc courant est 
nb_remaining_bytes_in_blk = SIZE_OF_SBLOCK - nro_current_byte_in_blk.


Si d.current_position + nb_bytes_to_read > d.size
nb_bytes_to_read = d.size - d.current_position.

Deux cas peuvent se prsenter :

CAS I  : nb_remaining_bytes_in_blk - nb_bytes_to_read < 0
- On copie les nb_remaining_bytes_in_block dans le buffer.
  Comment faire ?
  -- Faire un appel read_sblock(), on rcupre son dernier argument "a"
     de type sblock_t.
  -- On fait alors un memcopy dans le buffer  partir de la 
     position "a + nro_current_byte_in_blk" des 
     "nb_remaining_bytes_in_blk".
- On met  jour la position courante dans le fichier ouvert :
  d.current_position = d.current_position + nb_remaining_bytes_in_blk
- On diminue le nombre d'octets  lire :
  nb_bytes_to_read = nb_bytes_to_read - nb_remaining_bytes_in_blk
- On recommence avec les nouveaux arguments 
  -- d (a t modifi), 
  -- buffer + nb_remaining_bytes_in_blk,
  -- nb_bytes_to_read (a t modifi).
tant que nb_bytes_to_read > 0.
  
CAS II : nb_remaining_bytes_in_blk - nb_bytes_to_read >= 0
On copie nb_bytes_to_read dans le buffer.
  Comment faire ?
  -- Faire un appel read_sblock(), on rcupre son dernier argument "a"
     de type sblock_t.
  -- On fait alors un memcopy dans le buffer  partir de la 
     position "a + nro_current_byte_in_blk" de
     nb_bytes_to_read.
*/
error_t read_file(fd file_descr, void * buffer, int nb_bytes_to_read, int * p_nb_bytes_read)
{
int current_position;
int size_of_file;

int nro_current_block;
int nro_current_byte_in_block;
int nb_of_remaining_bytes_in_current_block;

struct s_fs_inode file_inode;
char data_block [SIZE_OF_SBLOCK];
error_t e;

void * buffer_start;           /* debug */
int    nb_bytes_to_read_start; /* debug */
  
  buffer_start = buffer;                     /* debug */
  nb_bytes_to_read_start = nb_bytes_to_read; /* debug */

  
  *p_nb_bytes_read = 0;

  
  /* DROIT DE LECTURE - AVORTER ? */
  if(file_descr->b_reading_authorized != 1)
    return ERROR_READING_NOT_AUTHORISED;
  
  
  /* AJUSTEMENT DU NOMBRE D'OCTETS A LIRE 
     Si l'on ne peut pas lire autant d'octets que demand 
     alors on ne lit que les octets disponibles, c'est--dire
     les octets  partir de la position courante, jusqu' la
     fin du fichier.
  */
  size_of_file = file_descr->size;
  
  if(file_descr->current_position + nb_bytes_to_read > size_of_file)
    nb_bytes_to_read = size_of_file - file_descr->current_position;
  
  /* LECTURE BLOC DE FICHIER */
  if((e = read_sblock(file_descr->nb_physical_block, (sblock_t) &file_inode)) != 0)
    return e;
  
  while(nb_bytes_to_read > 0)
  { 
    current_position = file_descr->current_position;
    nro_current_block = find_nro_current_block(file_descr);
    nro_current_byte_in_block = find_nro_current_byte_in_current_block(file_descr);
    nb_of_remaining_bytes_in_current_block = calc_number_of_remaining_bytes_in_current_block(file_descr);

    /* LECTURE BLOC DE DONNEES COURANT */
    if((e = read_sblock(big_endian_to_u16(file_inode.pf_data[nro_current_block]), (sblock_t) data_block)) != 0)
      return e;
 

    /* PREMIER CAS : le nombre d'octets restant  lire excde le
       nombre d'octets restant  lire dans le bloc courant, il
       faudra donc aller lire le bloc suivant */
    if(nb_of_remaining_bytes_in_current_block < nb_bytes_to_read) 
    {
      /* COPIE DES DONNEES DANS LE BUFFER */
      assert(buffer_start <= buffer &&
             nb_of_remaining_bytes_in_current_block >= 0 &&
             buffer + nb_of_remaining_bytes_in_current_block 
               <= buffer_start + nb_bytes_to_read_start); 
      
      memcpy(buffer, data_block + nro_current_byte_in_block, 
             nb_of_remaining_bytes_in_current_block);

      /* MISE A JOUR
         On met  jour la position courante dans le fichier ouvert et 
         le nombre d'octets restants  lire.
      */
      *p_nb_bytes_read = *p_nb_bytes_read + nb_of_remaining_bytes_in_current_block;
      current_position = current_position + nb_of_remaining_bytes_in_current_block;
      nb_bytes_to_read = nb_bytes_to_read - nb_of_remaining_bytes_in_current_block;
      buffer           = buffer           + nb_of_remaining_bytes_in_current_block;

      file_descr->current_position = current_position;
    }
    
    
    /* DEUXIEME CAS : le nombre d'octets  lire restant est
       infrieur ou gal au nombre d'octets restants  lire
       dans le bloc de donnes courant. */
    else /* nb_of_remaining_bytes_in_current_block >= nb_bytes_to_read */ 
    {
      /* COPIE DES DONNEES DANS LE BUFFER */
      assert(buffer_start <= buffer &&
             nro_current_byte_in_block >= 0 &&
             nro_current_byte_in_block + nb_bytes_to_read <= 256 &&
             buffer + nb_bytes_to_read <= buffer_start + nb_bytes_to_read_start);

      memcpy(buffer, data_block + nro_current_byte_in_block, nb_bytes_to_read);
      
      /* MISE A JOUR
         On met  jour la position courante dans le fichier ouvert et 
         le nombre d'octets restants  lire.
      */
      *p_nb_bytes_read = *p_nb_bytes_read + nb_bytes_to_read;
      buffer           = buffer           + nb_bytes_to_read;
      current_position = current_position + nb_bytes_to_read;
      nb_bytes_to_read = nb_bytes_to_read - nb_bytes_to_read;
      file_descr->current_position = current_position;

      assert(nb_bytes_to_read == 0);
    }
  }

  return 0;
}
/*-------------------------------------*/
/*             write_file              */
/*-------------------------------------*/
/*
Ecriture de "nb_bytes_to_write" octets dans un fichier
de descripteur de type fd "file_descr".


o Le fichier a une taille "file_descr->size".
o La position courante dans le fichier ouvert est 
  "file_descr->current_position".

o On se donne un buffer "buff_block_to_be_written" 
  (char [SIZE_OF_SBLOCK]) qui est destin  tre crit 
  physiquement aprs avoir t rempli.

  On le rinitialise  chaque boucle (bzero) 
  (cf boucle ci-dessous).

-- On se demande tout d'abord si on a le droit d'crire ou pas 
dans ce fichier (champ "file_descr->writing_authorized").

-- Tant que le nombre d'octets  crire est non nul :

---- On regarde si la position courante est confondue avec la 
taille du fichier :
(file_descr->current_position = file_descr->size)
et si "file_descr->size" est un multiple de SIZE_OF_SBLOCK.

  o Si c'est le cas, on a besoin d'allouer un nouveau bloc.
    Pour ce faire, on appelle successivement les fonctions :
    - find_new_data_block_number()
    - et alloc_new_data_block()
    On rcupre alors un numro de bloc : 
    "nro_physical_block_to_be_written"

  o Il faut galement rajouter ce nouveau bloc dans la liste des
    blocs de donnes du fichier (champ pf_data).
    Pour ce faire, il faut trouver l'indice du dernier bloc 
    "ind_last_block" et mettre "nro_physical_block_to_be_written"
    dans la case numro "ind_last_block" + 1 du champ pf_data du 
    fichier.
  
    Si (ind_last_block + 1 > 10) alors il faut retourner une erreur
    le tableau "pf_data" n'a que 10 cases (un block de fichier ne
    peut avoir plus de dix blocs de donnes associs).

---- Sinon,
  Le bloc  crire est celui dans lequel se trouve la position 
  courante (offset), c'est le bloc courant.
  Pour rcuprer son numro, on utilise find_nro_current_block().
  Son numro physique est le contenu de 
  pf_data[nro_block].

  On lit ce bloc de donnes dans le buffer "buff_block_to_be_written" :
  read_sblock(nro_physical_block_to_be_written, buff_block_to_be_written).

Dans tous les cas, on fait :

On copie, dans "buf_block_to_be_written",  la position retourne par
find_nro_current_byte_in_current_block(), le contenu de buffer,
nb_bytes_written = 
min(calc_number_of_remaining_bytes_in_current_block(), nb_bytes_to_be_written).

On crit physiquement le bloc ainsi obtenu :
write_sblock(nro_physical_block_to_be_written, buff_block_to_be_written).


On met  jour :
- La taille du fichier doit tre mise  jour
file_descr->size = max(file_descr->size, file_descr->current_position + nb_bytes_written)

- La position courante dans le fichier ouvert :
file_descr->current_position = file_descr->current_position + nb_bytes_written.
- Le nombre d'octets qu'il reste  crire :
nb_bytes_to_write = nb_bytes_to_write + nb_bytes_written
- L'adresse  laquelle se trouve le dbut de ce qu'il reste encore
 crire :
buffer = buffer + nb_bytes_written.
*/
error_t write_file(fd file_descr, void * buffer, int nb_bytes_to_write)
{
int file_size;
int current_position;
struct s_fs_inode inode;

int ind_last_block;
int nro_block_in_pf_data;

int nb_current_block;
int nb_physical_block_to_be_written;

char buff_block_to_be_written [SIZE_OF_SBLOCK];

int nb_current_byte_in_current_block;
int nb_of_remaining_bytes_in_block;

int nb_bytes_written;

int b_block_allocated;

error_t e;

  b_block_allocated = 0;
  
  file_size        = file_descr->size;
  current_position = file_descr->current_position;

  /* DROIT D'ECRITURE - AVORTER ? */
  if(file_descr->b_writing_authorized != 1)
    return ERROR_WRITING_NOT_AUTHORISED;
  
  /* LECTURE BLOC DE FICHIER */
  if((e = read_sblock(file_descr->nb_physical_block, (sblock_t) &inode)) != 0)
    return e;
  
  while(nb_bytes_to_write > 0)
  {
    /* INITIALISE "buff_block_to_be_written" avec
       "SIZE_OF_SBLOCK" zros */
    bzero(buff_block_to_be_written, SIZE_OF_SBLOCK);
    
    /* FIN DU FICHIER 
       Tous les blocs de donnes du fichier sont entirement remplis,
       il faut donc allouer un nouveau bloc de donnes pour le fichier */
    if((current_position == file_size) && ((file_size % SIZE_OF_SBLOCK) == 0))
    {
      /* INDICE DERNIER BLOC DE DONNEES 
         On recherche le nombre de blocs de donns associ au fichier 
         pour en dduire le premier indice libre pour un nouveau bloc
         de donnes.
      */
      ind_last_block = number_of_data_blocks_for_a_given_file(inode) - 1;
     
      /* FICHIER PLEIN - ABORT ? 
         Le fichier est plein, les dix blocs de donnes dont il
         dispose sont allous. NB_BLOCKS_MAX_PER_FILE = 10.
      */
      if(ind_last_block == NB_BLOCKS_MAX_PER_FILE - 1)
        return ERROR_FILE_FULL;

      /* INDICE NOUVEAU BLOC DE DONNEES 
         Indice du nouveau bloc de donnes dans "pf_data" */
      nro_block_in_pf_data = ind_last_block + 1;
      
      /* TROUVER UN NOUVEAU BLOC DE DONNEES LIBRE 
         On trouve un bloc libre, son numro (physique) est mis dans 
         "nb_physical_block_to_be_written" */
      if((e = find_new_data_block_number(&inode, &nb_physical_block_to_be_written)) != 0)
        return e;
     
      /* ALLOCATION NOUVEAU BLOC DE DONNEES
         On met  jour "tab_of_allocated_data_blocks_glob" */
      alloc_new_data_block(nb_physical_block_to_be_written); 
     
      /* NOUVEAU BLOC DE DONNEES POUR LE FICHIER
         On met  jour "pf_data" */
      inode.pf_data[nro_block_in_pf_data] = u16_to_big_endian(nb_physical_block_to_be_written);
     
      /* MARQUAGE : un nouveau bloc de donnes a t allou */
      b_block_allocated = 1;
    }

    /* LE DERNIER BLOC DE DONNEES DU FICHIER N'EST PAS PLEIN */
    else
    {
      /* RECUPERATION DU NRO PHYSIQUE DU DERNIER BLOC DE
         DONNEES DU FICHIER */
      nb_current_block = find_nro_current_block(file_descr);
      nb_physical_block_to_be_written = big_endian_to_u16(inode.pf_data[nb_current_block]);
      
      /* LECTURE DE CE BLOC */
      if((e = read_sblock(nb_physical_block_to_be_written, buff_block_to_be_written)) != 0)
        return e;
    }

    nb_current_byte_in_current_block = find_nro_current_byte_in_current_block(file_descr);
    nb_of_remaining_bytes_in_block   = SIZE_OF_SBLOCK - nb_current_byte_in_current_block;

    /* CALCULE NOMBRE D'OCTETS A ECRIRE DANS CETTE ITERATION */
    nb_bytes_written = min(nb_of_remaining_bytes_in_block, nb_bytes_to_write);
    
    /* ECRITURE DANS "buff_block_to_be_written" */
    memcpy(buff_block_to_be_written + nb_current_byte_in_current_block, buffer, nb_bytes_written);

    /* ECRITURE DU BLOC DE DONNEES */
    if((e = write_sblock(nb_physical_block_to_be_written, buff_block_to_be_written)) != 0)
    {
      if(b_block_allocated == 1)
        desalloc_data_block(nb_physical_block_to_be_written);
        
      write_sblock(file_descr->nb_physical_block, (sblock_t) &inode);
      
      return e;
    }
  
    /* MISE A JOUR 
       Mise  jour des nouveaux paramtres */
    file_size = max(file_size, current_position + nb_bytes_written);
    file_descr->size = file_size;
    inode.pf_size = u16_to_big_endian(file_size);
    
    current_position = current_position + nb_bytes_written;
    file_descr->current_position = current_position;
    
    nb_bytes_to_write = nb_bytes_to_write - nb_bytes_written;

    buffer = buffer + nb_bytes_written;
  }

  /* ECRITURE DU BLOC DE FICHIER */
  if((e = write_sblock(file_descr->nb_physical_block, (sblock_t) &inode)) != 0)
    return e;
  
  return 0;
}
/*----------------------------------------------------------*/
/*        number_of_data_blocks_for_a_given_file            */
/*----------------------------------------------------------*/
/*
Calcule le nombre de blocs de donnes associs  un bloc de fichier
 partir d'"inode".
*/
int number_of_data_blocks_for_a_given_file(struct s_fs_inode inode)
{
int nb;

  nb = big_endian_to_u16(inode.pf_size) / SIZE_OF_SBLOCK;

  if((big_endian_to_u16(inode.pf_size) % SIZE_OF_SBLOCK) != 0)
    return (nb + 1);

  else 
    return nb;
}
/*---------------------------------------------------*/
/*                find_nro_current_block             */
/*---------------------------------------------------*/
/*
Find current block given the current position.

Etant donne la position courante dans un fichier ouvert
(champ "current_position" d'une structure "struct s_file_descr")
cette fonction permet de calculer le numro de block courant
correspondant : "nro_current_block".

Le numro du bloc courant est :
nro_current_block = current_position / SIZE_OF_SBLOCK.

Par exemple :
------------

current_position est 33,
et les blocs sont de taille 10.

     0          1          2           3          4          5
_____________________________________________________________________
|          |          |          |   x      |          |          | ...
|__________|__________|__________|__________|__________|__________|___
     10         10         10         10         10         10


     
Ici, le bloc courant est le bloc numro (33 / 10) = 3.
C'est bien a.
*/
int find_nro_current_block(fd file_descr)
{
  return (file_descr->current_position / SIZE_OF_SBLOCK);
}
/*----------------------------------------------------------*/
/*           find_nro_current_byte_in_current_block         */
/* ---------------------------------------------------------*/
/*
Find the number of the current byte in the current block 
given the current position.

Etant donne la position courante dans un fichier ouvert
(champ "current_position" d'une structure "struct s_file_descr")
cette fonction permet de calculer le numro de l'octet courant
dans le bloc courant "nro_current_byte_in_current_block".

Le numro de l'octet courant dans le bloc courant est :
nro_current_byte_in_current_block = current_position % SIZE_OF_SBLOCK.

Par exemple :
------------

     0          1          2           3          4          5
                                     x
_____________________________________________________________________
|          |          |          |0123      |          |          | ...
|__________|__________|__________|__________|__________|__________|___
     10         10         10         10         10         10

Ici nro_current_byte_in_current_block = 33 % 10 = 3.
*/
int find_nro_current_byte_in_current_block(fd file_descr)
{
  return (file_descr->current_position % SIZE_OF_SBLOCK);
}
/*----------------------------------------------------------*/
/*     calc_number_of_remaining_bytes_in_current_block      */
/*----------------------------------------------------------*/
/*
Find the number of remaining bytes in block considering 
the number of the current byte in the current block.

Etant donn le numro de l'octet courant dans le block courant
cette fonction permet de calculer le nombre d'octets restants
dans le bloc courant  partir de ce numro d'octet jusqu' la 
fin du bloc.

Si l'on dispose du numro de l'octet courant dans le bloc 
courant (nro_current_byte_in_current_block) alors on en dduit 
le nombre d'octets restants dans le bloc courant  partir de 
cette position :
nb_of_remaining_bytes_in_curent_block =
SIZE_OF_BLOCK - nro_current_byte_in_current_block.
   
Par exemple :
------------

     0          1          2           3          4          5
                                     x
_____________________________________________________________________
|          |          |          |   1234567|          |          | ...
|__________|__________|__________|__________|__________|__________|___
     10         10         10         10         10         10

Le numro de l'octet courant dans le bloc courant est :
nro_current_byte_in_current_block = 3.
La taille du bloc est 10.

Donc le nombre d'octets restants dans le bloc courant 
(y compris l'octet courant) est 10 - 3 = 7.
*/
int calc_number_of_remaining_bytes_in_current_block(fd file_descr)
{
  return min((SIZE_OF_SBLOCK - find_nro_current_byte_in_current_block(file_descr)), 
             (file_descr->size - file_descr->current_position));
}
/*--------------------------------------------------*/
/*                   Minimum                        */ 
/*--------------------------------------------------*/
int min(int i1, int i2)
{
  if(i1 < i2)
    return i1;

  else
    return i2;
}
/*--------------------------------------------------*/
/*                   seek_file                      */ 
/*--------------------------------------------------*/
error_t seek_file(fd file_descr, int position)
{
  /* On ne peut "seeker" en-de ou au-del de la taille du
     fichier */
  if((position < 0) || (position > file_descr->size))
    return ERROR_SEEK;

  file_descr->current_position = position;

  return 0;
}
/*--------------------------------------------------*/
/*                   close_file                     */ 
/*--------------------------------------------------*/
void close_file(fd file_descr)
{
  free(file_descr);
}
/*--------------------------------------------------*/
/*                    stat_file                     */ 
/*--------------------------------------------------*/
/* Retourne des informations concernant le fichier dans une
   structure approprie (struct fs_stat). 
*/
error_t stat_file(fd file_descr, struct fs_stat * s_stat)
{
error_t e;
struct s_fs_inode inode;

  if((e = read_sblock(file_descr->nb_physical_block, (sblock_t) &inode)) != 0)
    return e;
  
  s_stat->st_protect = inode.pf_protect;
  s_stat->st_type    = inode.pf_type;
  s_stat->st_size    = big_endian_to_u16(inode.pf_size);
  s_stat->st_links   = big_endian_to_u16(inode.pf_links);
 
  return 0;
}
/*---------------------------------------------*/
/*             search_name_in_dir              */
/*---------------------------------------------*/
/*
Cette fonction permet de trouver le numro de bloc physique
d'un fichier de nom "name" qui se trouve  priori dans le 
rpertoire de numro de bloc physique "dir_physical_block_number".

En cas de succs, ce numro est plac dans la variable 
"p_physical_block_number" dont l'adresse est passe en argument.

En cas d'chec (entre autres : le rpertoire ne contient
pas de fichier de nom "name"), une erreur est retourne.
*/
error_t search_name_in_dir(char * name, 
                            int   dir_physical_block_number, 
                            int * p_physical_block_number)
{
       error_t  e;
            fd  file_descr_dir;
struct fs_stat  s_stat;
           int  nb_bytes_read;
     dir_entry  dir_entry_buffer;

     
  nb_bytes_read = -1;
  
  /* OUVERTURE FICHIER REPERTOIRE 
     qui a pour numro physique "dir_physical_block_number".
  */
  if((e = open_file_from_physical_block_number(dir_physical_block_number, 
                                               READING_AUTHORIZED,
                                               WRITING_PROHIBITED,
                                               WRITING_AT_THE_BEGINNING,
                                               &file_descr_dir)) != 0)
    return e;
    
  
  /* EST BIEN UN FICHIER REPERTOIRE ?
     On s'assure que l'argument "dir_physical_block_number"
     est bien le numro de block physique d'un fichier de
     type rpertoire (stat_file), sinon, on retourne une 
     erreur.
  */
  if((e = stat_file(file_descr_dir, &s_stat)) != 0)
  {  
    close_file(file_descr_dir);
    return e;
  }

  if(s_stat.st_type != TYPE_DIRECTORY)
  {  
    close_file(file_descr_dir);
    return ERROR_NOT_A_DIR;  
  }
   
  /* RECHERCHE DE "name" DANS LE REPERTOIRE */
  while(nb_bytes_read != 0)
  {
    /* LECTURE D'UNE ENTREE DE REPERTOIRE
       On lit dans le rpertoire une entre de rpertoire (de type dir_entry)
       de longueur sizeof(dir_entry_buffer).
       Le nombre d'octets lus est stok dans nb_bytes_read.
    */
    if ((e = read_file(file_descr_dir, &dir_entry_buffer, sizeof(dir_entry_buffer), &nb_bytes_read)) != 0)
    {  
      close_file(file_descr_dir);
      return e;
    }

    /* On vrifie que le nombre d'octets lus pour une entre
       est bien soit nul, soit gal  la taille d'une entre de rpertoire.
       Sinon, cela veut dire que l'entre est tronque et l'on retourne
       une erreur.
    */    
    if(nb_bytes_read == 0)
        break;
    
    if(nb_bytes_read != sizeof(dir_entry_buffer))
    {  
      close_file(file_descr_dir);
      return ERROR_DIRECTORY_ENTRY_TRUNCATED;
    }
      
    /* FICHIER DE NOM "name" DANS LE REPERTOIRE ?
       On compare le nom dans l'entre lue avec le nom "name" 
       que l'on recherche.
    */ 
    if(strcmp(name, dir_entry_buffer.filename) == 0)
    {
      /* FICHIER TROUVE - RECUPERATION DE SON NUMERO DE BLOC 
         On a trouv le nom "name" dans le rpertoire, donc on retourne son
         numro de bloc physique. */
      *p_physical_block_number = big_endian_to_u16(dir_entry_buffer.filenum);

      /* FERMETURE REPERTOIRE
         On ferme le fichier (rpertoire) que l'on a ouvert. */
      close_file(file_descr_dir);
      return 0;
    }
  }
  
  /* On ferme le fichier (rpertoire) que l'on a ouvert. */
  close_file(file_descr_dir);

  /* On retourne une erreur car on a pas trouv de fichier 
     portant ce nom dans le rpertoire. 
  */
  return ERROR_NO_SUCH_FILENAME_IN_DIR; 
}
/*-----------------------------------------------------------------*/
/*                      add_directory_entry                        */
/*-----------------------------------------------------------------*/
/*
Cette fonction permet de rajouter une entre (dir_entry) dans un 
rpertoire.

Elle prend un nom de fichier et un numro de bloc physique (numro
de bloc physique du fichier).

Elle prend aussi le numro de bloc physique du fichier rpertoire.


Remarque : on n'ajoute un nouveau fichier que lorsque l'on a dj 
son numro de bloc physique.
*/
error_t add_directory_entry(char * file_name, 
                             int   dir_physical_block_number, 
                             int   physical_block_number)
{
               fd  file_descr_dir;
        dir_entry  new_dir_entry;
struct s_fs_inode  inode;
   struct fs_stat  s_stat;
          error_t  e;

  /* NOM TROP LONG - ABORT ? 
     On vrifie que la longueur du nom du fichier n'est pas
     suprieure  30 octets (SIZE_MAX_FILE_NAMES). 
     Sinon, c'est une erreur. */
  if(strlen(file_name) >= 30)
    return ERROR_NAME_TOO_LONG;
  
  /* NOM DEJA DANS REPERTOIRE - ABORT ?
     On s'assure que ce nom n'existe pas dj dans le rpertoire. 
     Sinon, on retourne une erreur. */
  if(search_name_in_dir(file_name, dir_physical_block_number, &physical_block_number) == 0)
    return ERROR_NAME_ALREADY_EXISTS_IN_DIR;

  /* OUVERTURE DU REPERTOIRE
     On ouvre le fichier qui a pour numro physique
     "dir_physical_block_number", en criture, mode 
     "append at the end" */
  if((e = open_file_from_physical_block_number(dir_physical_block_number, 
                                               READING_PROHIBITED,
                                               WRITING_AUTHORIZED,
                                               WRITING_AT_THE_END,
                                               &file_descr_dir)) != 0)
    return e;
      
  /* EST BIEN UN FICHIER REPERTOIRE - CONTINUE ?
     On s'assure que l'argument "dir_physical_block_number"
     est bien le numro de block physique d'un fichier de
     type rpertoire (stat_file), sinon, on retourne une 
     erreur. */
  if((e = stat_file(file_descr_dir, &s_stat)) != 0)
  {  
    close_file(file_descr_dir);
    return e;
  }

  if(s_stat.st_type != TYPE_DIRECTORY)
  {  
    close_file(file_descr_dir);
    return ERROR_NOT_A_DIR;  
  }
 
  /* NOUVELLE ENTREE DE REPERTOIRE */
  new_dir_entry.filenum  = u16_to_big_endian(physical_block_number);
  strcpy(new_dir_entry.filename, file_name);

  /* LECTURE DU NOUVEAU BLOC DE FICHIER
     On doit maintenant incrmenter le nombre de liens dans
     le fichier que l'on vient d'ajouter.
     Pour ce faire, on doit lire l'inode correspondant au 
     fichier. */
  if((e = read_sblock(physical_block_number, (sblock_t) &inode)) != 0)
    return e;

  /* TROP DE LIENS - ABORT ?
  Le nombre de liens sur un fichier est cod sur deux octets (u16).
  Le nombre le plus grand que l'on puisse coder sur deux octets
  est NUMBER_MAX_CODED_ON_TWO_BYTES = (2^16 - 1) = 65.535.
  Donc, il faut s'assurer que le fichier n'a pas dj ce nombre de 
  liens. */
  if(big_endian_to_u16(inode.pf_links) == NUMBER_MAX_CODED_ON_TWO_BYTES)
    return ERROR_TOO_MANY_LINKS_ON_THE_SAME_FILE;
  
  /* AJOUTER UN LIEN
     Sinon, on peut mettre  jour le nombre de liens sur ce fichier,
     C'est--dire augmenter ce nombre d'une unit. */
  inode.pf_links = u16_to_big_endian(big_endian_to_u16(inode.pf_links) + 1);
  
  /* ECRITURE DE LA NOUVELLE ENTREE DANS LE REPERTOIRE 
     On crit cette nouvelle entre dans le rpertoire. 
     Elle est crite  la fin du rpertoire car on a ouvert
     le rpertoire en mode "append".
  */
  if((e = write_file(file_descr_dir, &new_dir_entry, SIZE_DIR_ENTRY)) != 0)
  {  
    close_file(file_descr_dir);
    return e;
  }

  /* On peut fermer le rpertoire. */
  close_file(file_descr_dir);
      
  /* ECRITURE DANS LE CACHE */
  if((e = write_sblock(physical_block_number, (sblock_t) &inode)) != 0)
    return e;

  return 0;
}
/*--------------------------------------------------*/
/*               remove_directory_entry             */
/*--------------------------------------------------*/
/*
Cette fonction permet de supprimer une entre dans un rpertoire.

Son principe est qu'elle remplace l'entre  supprimer par la
dernire entre du rpertoire.
*/
error_t remove_directory_entry(char * name, filenum_t dir_physical_block_number)
{
       error_t   e;
            fd   file_descr_dir;
struct fs_stat   s_stat;
           int   dir_size;
           int   nb_of_entries;
     dir_entry * tab_of_entries;
           int   nb_bytes_read;
           int   i;
           int   physical_bloc_number_of_the_suppressed_entry;

  /* OUVERTURE DU REPERTOIRE EN LECTURE-ECRITURE
     On commence par ouvrir le fichier de numro de bloc physique 
     "dir_physical_block_number" en lecture et en criture. */
  if((e = open_file_from_physical_block_number(dir_physical_block_number, 
                                               READING_AUTHORIZED,
                                               WRITING_AUTHORIZED,
                                               WRITING_AT_THE_BEGINNING,
                                               &file_descr_dir)) != 0)
    return e;

  /* EST BIEN UN FICHIER REPERTOIRE - CONTINUE ?
     On s'assure que l'argument "dir_physical_block_number"
     est bien le numro de block physique d'un fichier de
     type rpertoire (stat_file), sinon, on retourne une 
     erreur.
  */
  if((e = stat_file(file_descr_dir, &s_stat)) != 0)
  {  
    close_file(file_descr_dir);
    return e;
  }
    
  if(s_stat.st_type != TYPE_DIRECTORY)
  {  
    close_file(file_descr_dir);
    return ERROR_NOT_A_DIR;  
  }
  
  /* NOMBRE D'ENTREES DE REPERTOIRE
     On cherche le nombre d'entres dans le rpertoire. */
  dir_size      = s_stat.st_size;
  nb_of_entries = dir_size / SIZE_DIR_ENTRY; 
  
  /* TABLEAU POUR STOCKER TOUTES LES ENTREES 
     On alloue un tableau pour stocker toutes les entres du
     rpertoire. */
  tab_of_entries = (dir_entry *) malloc(sizeof(dir_entry)*nb_of_entries);

  if(tab_of_entries == NULL)
  {  
    close_file(file_descr_dir);
    return ERROR_MALLOC;
  }
  
  /* LECTURE DU REPERTOIRE
     On lit tout le rpertoire que l'on rcupre dans "tab_of_entries" */
  if((e = read_file(file_descr_dir, 
                    tab_of_entries, 
                    sizeof(dir_entry)*nb_of_entries,
                    &nb_bytes_read))        != 0)
  {  
    free(tab_of_entries);
    close_file(file_descr_dir);
    return e;
  }
   
  /* ENTREE A SUPPRIMER DANS LE REPERTOIRE
     On recherche dans le tableau des entres,
     l'entre que l'on veut supprimer,
     on rcupre  la sortie de la boucle son indice
     dans le tableau.
  */
  for(i=0 ; i<nb_of_entries ; i++)
  {
    if(strcmp(name, tab_of_entries[i].filename) == 0)
      break;
  }

  /* PAS DE TELLE ENTREE DANS LE REPERTOIRE - ABORT ?
     Dans ce cas, le file n'est pas dans le rpertoire. */
  if(i == nb_of_entries)
  {  
    free(tab_of_entries);
    close_file(file_descr_dir);
    return ERROR_NO_SUCH_FILENAME_IN_DIR;
  }

  /* MISE A JOUR BLOC DE FICHIER
     Je garde en mmoire le numro du bloc physique correspondant
      l'entre que je veux supprimer pour ensuite mettre  jour
     le nombre de liens sur le fichier correspondant et
     ventuellement le supprimer. */
  physical_bloc_number_of_the_suppressed_entry = big_endian_to_u16(tab_of_entries[i].filenum);

  /* on libre l'inode */
  if((e = decrease_number_of_links_and_suppress_file_if_necessary(
              physical_bloc_number_of_the_suppressed_entry)) != 0)
  {  
    free(tab_of_entries);
    close_file(file_descr_dir);
    return e;
  }
  
  /* MISE A JOUR BLOC DE FICHIER REPERTOIRE
     On remplace l'entre que l'on veut supprimer par la 
     dernire entre du rpertoire */
  tab_of_entries[i] = tab_of_entries[nb_of_entries - 1];    
  
  /* SEEKER AU DEBUT DU REPERTOIRE 
     On crit physiquement dans le rpertoire la modif.
     Mais avant cela il faut se replacer au dbut du fichier
     rpertoire car le read_file ci-dessus a dplac la position 
     courante dans le fichier ouvert. */
  if((e = seek_file(file_descr_dir, WRITING_AT_THE_BEGINNING)) != 0 )
  {  
    free(tab_of_entries);
    close_file(file_descr_dir);
    return e;
  }
  
  /* ECRITURE DU REPERTOIRE
     Le "-1" car on vient d'enlever une entre. */
  if((e = write_file(file_descr_dir, tab_of_entries, (nb_of_entries - 1) * SIZE_DIR_ENTRY)) != 0)
  {  
    free(tab_of_entries);
    close_file(file_descr_dir);
    return e;
  }

  /* MISE A JOUR TAILLE DU REPERTOIRE ET DES TABLEAUX DES BLOCS
     LIBRES */
  if((e = truncate_file(file_descr_dir, file_descr_dir->size - SIZE_DIR_ENTRY)) != 0)
  {  
    free(tab_of_entries);
    close_file(file_descr_dir);
    return e;
  }

  free(tab_of_entries);
  close_file(file_descr_dir);
  return 0;
}
/*----------------------------------------*/  
/*                readdir_entry           */
/*----------------------------------------*/  
/*
Cette fonction permet de lire une entre dans un rpertoire
 la position courante dans le fichier ouvert correspondant.

Fonction appele dans "couche_applications.c".

Args
````
    file_descr : le descripteur de fichier ouvert

          p_de : dans laquelle on rcupre l'entre de rpertoire

p_b_end_of_dir : boolen qui vaut 1 si on est arriv  la fin
                 du rpertoire, 0 sinon.
*/
error_t readdir_entry(fd file_descr, dir_entry * p_de, int * p_b_end_of_dir)
{
       error_t  e;
           int  nb_bytes_read;
struct fs_stat  s_stat;

  /* EST BIEN UN FICHIER REPERTOIRE - CONTINUE ?
     On s'assure que le descripteur de fichier est celui d'un 
     rpertoire. */   
  if((e = stat_file(file_descr, &s_stat)) != 0)
    return e;
    
  if(s_stat.st_type != TYPE_DIRECTORY)
    return ERROR_NOT_A_DIR;  

  /* LECTURE D'UNE ENTREE DE REPERTOIRE 
     On lit dans ce fichier ouvert une entre de rpertoire
     ( la position courante).
     (La position courante dans le fichier ouvert est mise  jour
     dans "read_file").
     On rcupre cette entre dans "p_de" pass en argument  la 
     fois  la fonction "read_file" et  la fonction elle-mme.
  */
  if((e = read_file(file_descr, p_de, SIZE_DIR_ENTRY, &nb_bytes_read)) != 0)
    return e;
   
  /* FIN DE REPERTOIRE ATTEINTE */
  if(nb_bytes_read == 0)
    *p_b_end_of_dir = 1;

  /* ENTREE DE REPERTOIRE TRONQUEE - ABORT */
  else if(nb_bytes_read != SIZE_DIR_ENTRY)
    return ERROR_DIRECTORY_ENTRY_TRUNCATED; 
  
  /* FIN DU REPERTOIRE PAS ENCORE ATTEINTE */
  else 
    *p_b_end_of_dir = 0;

  return 0;
}
/*----------------------------------------------------------*/
/*                       truncate_file                      */
/*----------------------------------------------------------*/
/*
Cette fonction permet de tronquer un fichier  new_size.
L'ancienne taille devient la nouvelle.

Elle met galement  jour les deux tableaux 
"tab_of_allocated_file_blocks_glob" et
"tab_of_allocated_data_blocks_glob".
*/
error_t truncate_file(fd file_descr, int new_size)
{
struct s_fs_inode  inode;
          error_t  e;

  /* NOUVELLE TAILLE PAS REGLEMENTAIRE - ABORT ?
     On doit s'assurer que la nouvelle taille n'est pas 
     suprieure  l'actuelle (pas autoris). 
     On doit aller lire la taille actuelle. */
  /* LECTURE BLOC DE FICHIER 
     Pour rcuprer sa taille. */          
  if((e = read_sblock(file_descr->nb_physical_block, (sblock_t) &inode)) != 0)
    return e;
  
  if(big_endian_to_u16(inode.pf_size) < new_size)
    return ERROR_TRUNCATE;
 
  /* MISE A JOUR NOUVELLE TAILLE
     On met  jour la taille du fichier dans l'inode et dans 
     le descripteur. */
  inode.pf_size   = u16_to_big_endian(new_size);
  file_descr->size = new_size;
  
  /* ECRITURE DU REPERTOIRE
     Enfin, on met  jour physiquement le rpertoire. */
  if((e = write_sblock(file_descr->nb_physical_block, (sblock_t) &inode)) != 0)
    return e;
 
  /* MISE A JOUR DES TABLEAUX DES BLOCS LIBRES
     On met  jour les tableaux des blocs libres de fichier 
     et de donnes, ceci, car en tronquant des fichiers
     on a pu librer des blocs. */
  if((e = update_tabs_of_free_blocks()) != 0)
    return e;

  return 0;
}
/*---------------------------------------------------------*/
/* decrease_number_of_links_and_suppress_file_if_necessary */
/*---------------------------------------------------------*/
/*
Cette fonction permet de supprimer un lien sur un fichier
et ventuellement de supprimer le fichier si ce lien 
tait le dernier.
*/
error_t decrease_number_of_links_and_suppress_file_if_necessary(
                                filenum_t physical_block_number)
{
struct s_fs_inode  inode;
          error_t  e;

  /* LECTURE BLOC DE FICHIER */          
  if((e = read_sblock(physical_block_number, (sblock_t) &inode)) != 0)
    return e;

  /* FICHIER REPERTOIRE - PAS SUPPRIMER LIEN ?
     Si le fichier est un rpertoire non vide alors il n'a qu'un 
     seul lien et il ne faut pas le supprimer. */
  if((inode.pf_type == TYPE_DIRECTORY) && (inode.pf_size != 0))
    return ERROR_DIRECTORY_NOT_EMPTY;
    
  /* UN SEUL LIEN
     Le lien que l'on veut supprimer est le dernier. */
  if(big_endian_to_u16(inode.pf_links) == 1)
  {
    /* Le fichier est libr */
    inode.pf_state = u16_to_big_endian(STATE_FREE);

    /* Ecriture du bloc dans le cache */
    if((e = write_sblock(physical_block_number, (sblock_t) &inode)) != 0)
      return e;
    
    /* On met  jour la table des blocs libres. */
    if((e = update_tabs_of_free_blocks()) != 0)
      return e;
  }

  /* PLUSIEURS LIENS
     Il reste encore au moins deux liens sur ce fichier. 
     On dcrmente d'une unit le nombre de liens sur ce fichier.
     On crit le bloc dans le cache. */
  else
    inode.pf_links = u16_to_big_endian(big_endian_to_u16(inode.pf_links) - 1);

  if((e = write_sblock(physical_block_number, (sblock_t) &inode)) != 0)
    return e;
  
  return 0;
}
/*------------------------------------------------*/
/*         return_head_and_tail_of_path           */
/*------------------------------------------------*/
/*
Cette fonction permet de retourner un pointeur sur la 
tte d'un chemin et un sur le reste.

Un chemin est chane de caractres :
/r1/r2/r3/r4/...

La tte du chemin est r1
Le reste du chemin est /r2/r3/r4/...

La fonction prend le chemin en argument.

Les pointeurs sur la tte et sur le reste sont mis--jour
dans les arguments de la fonction :
p_head et p_tail.
*/
error_t return_head_and_tail_of_path(char * path, char ** p_head, char ** p_tail)
{ 
 int   i;
 int   j;
char * head;

  /* PAS CHEMIN ABSOLU - ABORT ?
     Si le "path" ne commence pas par un '/' alors on retourne une erreur.
     Il faut un chemin absolu. */
  if(path[0] != '/')
    return ERROR_NOT_AN_ABSOLUTE_PATH;
  
  /* Le premier caractre de "path" est '/' donc on passe
     directement au deuxime. */
  i=1;

  /* INDICE NEXT OCCURENCE OF '/'
     On recherche l'indice dans "path" du premier '/'. */
  while((path[i] != '/') && (path[i] != '\0'))
    i++;
  
  /* ALLOCATION MEMOIRE POUR LA TETE DU PATH
     On fabrique une petite chane de caractres destine
      contenir le premier mot (sans slashs) de "path". */
  head = (char *) malloc(sizeof(char)*i);

  if(head == NULL)
    return ERROR_MALLOC;
  
  /* COPIE DE LA TETE DE "path" DANS "head" */ 
  for(j=0 ; j<i-1 ; j++)
    head[j] = path[j+1];
    
  head[j] = '\0';
  
  *p_head = head;
  
  /* TAIL 
     D'o "p_tail"
     Remarque : **p_tail vaut soit '/' soit '\0'. */ 
   *p_tail = path + i;

  return 0;
}
/*--------------------------------------------------------*/
/*             path_to_physical_block_number              */
/*--------------------------------------------------------*/
/*
Cette fonction permet, tant donn un chemin, de retourner
le numro de bloc physique correspondant au fichier que ce
chemin dsigne.

D'autre part, si ce fichier n'existe pas (mais que le chemin
est correct jusqu'au nom de fichier), on retourne le numro 
de bloc physique du rpertoire que ce chemin dsigne.

Args 
````                  path : le chemin

   p_physical_block_number : dans lequel on rcupre le numro
                             de bloc.
                               
b_file_not_already_created :  1 si le fichier n'existe pas
                             dj,  zro sinon.
                               
          p_last_component : dans lequel on rcupre le nom du 
                             fichier que "path" dsigne si ce 
                             fichier n'existe pas sinon, 
                             ce pointeur vaut NULL.
*/
error_t path_to_physical_block_number(char *  path, 
                                       int *  p_physical_block_number,
                                       int    b_file_not_already_created,
                                      char ** p_last_component
                                      )
{
error_t   e;
    int   current_dir_physical_block_number;
   char * current_path;
   char * head;
   char * tail;

  /* DEPART : REPERTOIRE RACINE '/'
     Initialement, on est dans le rpertoire "racine"
     qui a pour numro de bloc physique 1 (super-bloc). */
  current_dir_physical_block_number = 1;
  
  /* DEPART : CHEMIN COURANT "path"
     Initialement, le chemin courant est "path". */
  current_path = path;

  /* LE CHEMIN DESIGNE LE REPERTOIRE RACINE */
  if((path[0] == '/') && (path[1] == '\0') & (b_file_not_already_created == 0))
  {
    * p_physical_block_number = 1; /* superblock */
    return 0;
  }

  do
  {
    /* HEAD AND TAIL OF PATH */
    if((e = return_head_and_tail_of_path(current_path, &head, &tail)) != 0)
      return e; 

    /* TAIL EST VIDE ET LE FICHIER DE NOM "head" N'EXISTE PAS 
       On est arriv  la fin du parcours du "path" et le flag "fichier pas
       encore cr" est  1. */
    if((tail[0] == '\0') && (b_file_not_already_created == 1))
    {
      *p_last_component = head;
      break;
    }

    /* TAIL EST VIDE ET LE FICHIER DE NOM "head" EXISTE DANS LE
       REPERTOIRE COURANT
       Sinon, on recherche le fichier dans le rpertoire courant
       et l'argument "p_physical_block_number" prend sa valeur. */
    if((e = search_name_in_dir(head, current_dir_physical_block_number, p_physical_block_number)) != 0)
      return e;
      
    /* TAIL N'EST PAS VIDE
       On met  jour les paramtres "current_dir_physical_block_number"
       et "current_path" pour continuer  descendre dans le "path". */
    current_dir_physical_block_number = *p_physical_block_number;
    current_path = tail;
    free(head);
  }
  while(tail[0] != '\0');

  /* FICHIER N'EXISTE PAS ENCORE DANS LE REPERTOIRE COURANT 
     On retourne donc le numro de bloc du rpertoire courant. */
  if(b_file_not_already_created == 1)
    *p_physical_block_number = current_dir_physical_block_number; 
 
  return 0;
}
/*--------------------------------------*/
/*              open_file               */
/*--------------------------------------*/
/*
Cette fonction permet d'ouvrir un fichier 
 partir de son nom et ventuellement de
le crer s'il n'existe pas dj.

Rcupre un descripteur de fichier pour ce fichier.

Args
````
        name : rfrence absolue du fichier.
     
    b_create : 1 si le fichier n'existe pas dj
               sinon le fichier existe dj.

 b_directory : 1 si le fichier  crer est un rpertoire
               sinon le fichier  crer est rgulier

p_file_descr : descripteur de fichier, variable dans laquelle 
               on rcupre un descripteur de fichier pour le 
               fichier.
*/
error_t open_file(char * name,
                   int   b_reading_authorized, 
                   int   b_writing_authorized, 
                   int   b_writing_at_the_end,
                   int   b_create,
                   int   b_directory,
                    fd * p_file_descr)
{
error_t   e;  
    int   dir_physical_block_number;
    int   type;
    int   permissions;
    int   file_block_number;
   char * file_name_to_be_created;

  /*------ FICHIER PAS ENCORE CREE ------
     "name" est un chemin absolu (expl. /r1/r2/r3/.../titi) 
     dsignant le fichier de nom "titi" dans le rpertoire 
     /r1/r2/r3/... et "titi" n'existe pas encore. */
  if(b_create == 1)
  {
    /* RECUPERATION DU NUMERO DE BLOC DU REPERTOIRE
       /r1/r2/r3/... DANS "dir_physical_block_number"
       ET DU NOM DE FICHIER "titi" dans "file_name_to_be_created" 
       
       Arguments :
       "1" : pour dire que le fichier n'a pas encore t cr
       "file_name_to_be_created" : pour rcuprer le nom du fichier
                                    crer. 
    */
    if((e = path_to_physical_block_number(name, 
                                          &dir_physical_block_number, 
                                          1, 
                                          &file_name_to_be_created)) != 0) 
      return e;

    /* TYPE FICHIER A CREER : REPERTOIRE
       On dtermine si le fichier  crer est un rpertoire ou pas. */
    if(b_directory == 1)
      type = TYPE_DIRECTORY;

    /* TYPE FICHIER A CREER : REGULIER */
    else
      type = TYPE_REGULAR_FILE;
      
    /* PERMISSIONS FICHIER A CREER
       On calcule les permissions. */
    permissions = decrypt_permissions(b_writing_authorized, b_reading_authorized);
    
    /* CREATION DU FICHIER - RECUPERATION DE SON NUMERO DE BLOC
       Son numro de bloc attribu est plac dans la 
       variable "file_block_number".
    */
    if((e = create_file(type, permissions, &file_block_number)) != 0) 
    {
      free(file_name_to_be_created);
      return e;
    }
    
    /* NOUVELLE ENTREE POUR LE NOUVEAU FICHIER DANS LE
       REPERTOIRE COURANT */
    if((e = add_directory_entry(file_name_to_be_created, 
                                dir_physical_block_number, 
                                file_block_number)) != 0)
    {
      free(file_name_to_be_created);
      return e;
    }
    
    /* (Le malloc a t fait dans l'appel 
       "path_to_physical_block_number".) */
    free(file_name_to_be_created);
    
    /* OUVERTURE DU FICHIER A PARTIR DE SON NUMERO DE BLOC
       Enfin, on peut ouvrir le fichier. 
       Le descripteur de fichier est plac dans "p_file_descr"
       et rendu  la fonction courante.
    */  
    if((e = open_file_from_physical_block_number(file_block_number,   
                                                 b_reading_authorized, 
                                                 b_writing_authorized, 
                                                 b_writing_at_the_end,
                                                 p_file_descr)) != 0)
      return e;

    return 0;
  }
  
  /*------- FICHIER EXISTE DEJA -------*/
  else
  {
    /* RECUPERATION DU NUMERO DE BLOC DU FICHIER
       On retourne le numro de bloc physique du fichier  ouvrir
       dans "physical_block_number".
       L'argument "0" pour dire que le fichier a dj t cr.
       L'argument "NULL" car il n'y a rien  crer. */
    if((e = path_to_physical_block_number(name, &file_block_number, 0, NULL)) != 0) 
      return e;
    
    /* OUVERTURE DU FICHIER A PARTIR DE SON NUMERO DE BLOC
       Enfin, on peut ouvrir le fichier. 
       Le descripteur de fichier est plac dans "p_file_descr"
       et rendu  la fonction courante. */  
    if((e = open_file_from_physical_block_number(file_block_number,   
                                                 b_reading_authorized, 
                                                 b_writing_authorized, 
                                                 b_writing_at_the_end,
                                                 p_file_descr)) != 0)
      return e;
  }

  return 0;
}
/*--------------------------------------*/
/*          decrypt_permissions         */
/*--------------------------------------*/
/*
Il existe trois types de permissions :
- aucune permission                           0 
- permission d'criture seule (PERMIT_WRITE)  1
- permission de lecture seule (PERMIT_READ)   2
- permission de lecture et d'criture         3
*/
int decrypt_permissions(int b_writing_authorized, int b_reading_authorized)
{
  int permissions;

  permissions = 0;

  if(b_writing_authorized != 0) 
    permissions = permissions | PERMIT_WRITE;

  if(b_reading_authorized != 0) 
    permissions = permissions | PERMIT_READ;
  
  return permissions;
}
/*--------------------------------------*/
/*              link_file               */
/*--------------------------------------*/
/*
Cette fonction permet de faire un lien sur un fichier.
*/
error_t link_file(char * old_path, char * new_path)
{
error_t   e;
    int   file_block_number;
    int   dir_physical_block_number;
   char * new_name;

  /* RECUPERATION NRO DE BLOC DU FICHIER 
     "old_path" dsigne le nom du fichier sur lequel 
     on veut faire un lien.
     Args : "0" car le fichier existe 
            "NULL" pour la mme raison */  
  if((e = path_to_physical_block_number(old_path, &file_block_number, 0, NULL)) != 0) 
    return e;
  
  /* RECUPERATION DU NRO DE BLOC DU REPERTOIRE DANS LEQUEL ON VEUT
     METTRE LE NOUVEAU LIEN (dans "dir_physical_block_number")
     ET DU NOM DU NOUVEAU LIEN (dans "new_name") */ 
  if((e = path_to_physical_block_number(new_path, &dir_physical_block_number, 1, &new_name)) != 0) 
    return e;

  /* NOUVELLE ENTREE POUR LE NOUVEAU LIEN */
  if((e = add_directory_entry(new_name, dir_physical_block_number, file_block_number)) != 0)
  {
    free(new_name);
    return e;
  }
  
  free(new_name);
  return 0;
}
/*--------------------------------------------------*/
/*              Test read, write, ...               */ 
/*--------------------------------------------------*/

char chiffres[] = "0123456789";

int test_read_write_logiques(int size_max_file, int nb_of_tests)
{
char * tampon_ecriture;
char * tampon_lecture;
int current_size_file; 
error_t e;
int nb_physical_block_file;
fd file_descr;
int type_of_operation;
int i;
int j;
int origine_segment;
int extremite_segment;
int nb_bytes_to_write;
int nb_bytes_to_read;
unsigned char * adresse_demarrage_lecture_dans_tampon_ecriture;
int nb_bytes_read;

  current_size_file = 0; 
 
  tampon_ecriture = (unsigned char*) malloc(sizeof(unsigned char)*size_max_file);
  tampon_lecture  = (unsigned char*) malloc(sizeof(unsigned char)*size_max_file);

  /* On cr et ouvre un fichier. */
  if((e = create_file(TYPE_REGULAR_FILE, PERMIT_READ | PERMIT_WRITE, &nb_physical_block_file)) != 0)
  {
    printf("Error in create_file : %s\n", string_of_error(e));
    return -1;
  }
  
  if((e = open_file_from_physical_block_number(nb_physical_block_file, 1, 1, 1, &file_descr)) != 0)
  {
    printf("Error in open_file_from_physical_block_number : %s\n", string_of_error(e));
    return -1;
  }
  
  for(i=0 ; i<nb_of_tests ; i++)
  {
    /* On tire un entier au hasard entre 0 et 1
       - Si on a tir 1 alors on fait une criture,
       - sinon, on fait une lecture.
    */
    type_of_operation = rand() % 2;

    if(type_of_operation == 1) /* On fait une riture. */
    {
      /* On choisit un segment du tableau tampon_ecriture. */  
      
      /* On choisit au hasard l'origine du segment, un entier compris 
         entre 0 et la position courante dans le fichier ouvert.
      */

      /* Si la taille courante du fichier n'est pas gale  la taille
       * maximale du fichier, on s'autorise  prendre comme origine du
       * segment la fin du fichier (car il y a encore de la place) ;
       * sinon l'origine doit tre prise entre 0 et size_max_file - 1. */
      if(current_size_file == size_max_file)
        origine_segment = rand() % current_size_file;

      else
        origine_segment = rand() % (current_size_file + 1);
      
      /* On choisit au hasard l'extrmit du segment, origine_segment +
         un entier compris entre 0 et size_max_file - origine_segment.      
      */
      extremite_segment = origine_segment + (rand() % (size_max_file - origine_segment));

      /* On remplit les octets origine_segment  extremite_segment 
         du tampon_ecriture.
      */
      for(j=origine_segment ; j<=extremite_segment ; j++)
        tampon_ecriture[j] = chiffres[rand() % 10]; /* Caractres : entiers compris entre 0 et 255 */

      /* Dans le fichier, maintenant, on "seek"  la position origine_segment. */
      if((e = seek_file(file_descr, origine_segment)) != 0)
      {
        printf("Error in seek_file : %s\n", string_of_error(e));
        return -1;
      }

      /* On crit maintenant dans le fichier l'quivalent du contenu 
         du tampon_ecriture de la position "origine_segment"  la position
         "extremite_segment" soit "nb_bytes_to_write".
      */
      nb_bytes_to_write = extremite_segment - origine_segment + 1;
      printf("test:write or=%d ex=%d\n", origine_segment, extremite_segment);
      fflush(stdout);
  
      if((e = write_file(file_descr, tampon_ecriture + origine_segment , nb_bytes_to_write) ) != 0)
      {
        printf("Error in write_file : %s\n", string_of_error(e));
        return -1;
      }

      current_size_file = max(current_size_file, origine_segment + nb_bytes_to_write);
    }

    else /* On fait une lecture. */
    {
      /* On choisit un segment  lire dans le tampon_de_lecture. */

      origine_segment   = rand() % current_size_file;
      /* Ici current_size_file - origine_segment ne peut tre nul car origine_segment est
       * compris entre 0 et current_size_file - 1. */
      extremite_segment = origine_segment + (rand() % (current_size_file - origine_segment));

      printf("test:read or=%d ex=%d\n", origine_segment, extremite_segment);
      fflush(stdout);

      /* On va aller lire dans le fichier le mme segment. */
      
      /* On fait un seek dans le fichier  la position origine_segment. */
      if((e = seek_file(file_descr, origine_segment)) != 0)
      {
        printf("Error in seek_file : %s\n", string_of_error(e));
        return -1;
      }

      /* On lit le segment correspondant dans le fichier. */ 
      nb_bytes_to_read = extremite_segment - origine_segment + 1;
      
      if((e = read_file(file_descr, tampon_lecture, nb_bytes_to_read, &nb_bytes_read)) != 0)
      {
        printf("Error in read_file : %s\n", string_of_error(e));
        return -1;
      }
      
      if(nb_bytes_to_read != nb_bytes_read)
      {
        printf("Expected %d, got %d.\n", nb_bytes_to_read, nb_bytes_read);  
        return -1;
      }

      /* On compare octet  octet les contenus du tampon_lecture et du fichier */
      adresse_demarrage_lecture_dans_tampon_ecriture = tampon_ecriture + origine_segment; 
      
      for(j=0 ; j<nb_bytes_to_read ; j++)
      {
        if(tampon_lecture[j] != *(adresse_demarrage_lecture_dans_tampon_ecriture + j))
        {
          printf("Data error at position %d\n", j + origine_segment);
          return -1;
        }
      }
    }
  }
  return 0;
}
