/* Auteur - La Massiot - 2005 */

/*____________________________________*/
/*------------------------------------*/
/*          couche_logique.h          */
/*____________________________________*/
/*------------------------------------*/

#include "types.h"
#include "cache.h"




#define          PD_FREE_SPACE  240  /* PD : Partition de description */
#define          PF_FREE_SPACE  224  /* PF : Partition des fichiers   */
#define NB_BLOCKS_MAX_PER_FILE   10
#define      TYPE_REGULAR_FILE    0
#define         TYPE_DIRECTORY    1
#define             STATE_FREE    0
#define         STATE_OCCUPIED    1
#define            PERMIT_READ    1
#define           PERMIT_WRITE    2

#define           SIZE_MAX_FILE_NAMES    30
#define            READING_AUTHORIZED     1
#define            READING_PROHIBITED     0
#define            WRITING_AUTHORIZED     1
#define            WRITING_PROHIBITED     0
#define            WRITING_AT_THE_END     1
#define      WRITING_AT_THE_BEGINNING     0
#define                SIZE_DIR_ENTRY    32
#define NUMBER_MAX_CODED_ON_TWO_BYTES 65535 

#define      SUPER_BLOCK_BLOCK_NUMBER     1
#define            DOESNT_MATTER_HERE     0
#define                   CREATE_FILE     1
#define            DO_NOT_CREATE_FILE     0
#define                     NOT_A_DIR     0
#define                         A_DIR     1
#define      FILE_NOT_ALREADY_CREATED     1
#define          FILE_ALREADY_CREATED     0

/*-------------------------------------------------------------------------*/
/*                          Partition de description                       */
/*-------------------------------------------------------------------------*/
/* 
   -- pd pour "Partition de Description" 
   -- n est le nombre de fichiers demands par l'utilisateur 
      lorsqu'il construit son disque logique vide prt  l'emploi. 
   -- d est le nombre de blocks de donnes demands par l'utilisateur 
      lorsqu'il construit son disque logique vide prt  l'emploi.
   -- A ce niveau les blocs sont de taille 256 octets.
   -- free space : 4+4+4+4 = 16 
      256-16  = 240 
*/
/*-------------------------------------------------------------------------*/
struct s_partition_de_description
{
  u32  pd_n;                          /* n                           */
  u32  pd_nb_of_free_files;           /* n au dpart                 */
  u32  pd_nb_of_free_blocks;          /* d au dpart                 */
  u32  pd_nro_first_free_block;       /* le n plus-unime au dpart  */
  char pd_free_space [PD_FREE_SPACE];
};




/*--------------------------------------------------------------*/
/*                      Partition des fichiers                  */
/*--------------------------------------------------------------*/
/*
   -- fs pour "File System" 
   -- pf pour "Partition des Fichiers"
   -- free space : 32 + 4*16 + 2*8 = 112
      112/8 = 14
      256 - 14 = 242
*/
/*-------------------------------------------------------------*/
struct s_fs_inode
{
  u8  pf_protect;
  u8  pf_type;
  u16 pf_size;
  u16 pf_data [NB_BLOCKS_MAX_PER_FILE];
  u16 pf_state;                          /* Au dbut :  0 car il est libre */
  u16 pf_links;
  u32 pf_reserved;
  char pf_free_space [PF_FREE_SPACE];
};



/*oooooooooooooooooooooooooooooooooooooooooooooooo*/
/* Structure file descriptor */
struct s_file_descr
{
  int nb_physical_block;
  int size;
  int b_reading_authorized;
  int b_writing_authorized;
  int current_position;
};

typedef struct s_file_descr * fd;
/*oooooooooooooooooooooooooooooooooooooooooooooooo*/
struct fs_stat
{
  u8  st_protect;
  u8  st_type;
  u16 st_size;
  u16 st_links; 
};
/*oooooooooooooooooooooooooooooooooooooooooooooooo*/
typedef u16 filenum_t;

struct s_dir_entry
{
  filenum_t filenum;
  char filename [SIZE_MAX_FILE_NAMES];
};

typedef struct s_dir_entry dir_entry;
/*oooooooooooooooooooooooooooooooooooooooooooooooo*/

/*---------- FONCTIONS -----------*/

u16 u16_to_big_endian(u16 a);
u32 u32_to_big_endian(u32 a);
u16 big_endian_to_u16(u16 a);
u32 big_endian_to_u32(u32 a);

error_t init_couche_logique();

error_t shutdown_couche_logique();

error_t init_tabs_of_free_blocks();
error_t update_tabs_of_free_blocks();

int nb_blocks_given_the_size_of_a_file(int size);

error_t find_new_file_block_number(int * p_result);
error_t find_new_data_block_number(struct s_fs_inode * p_inode, 
                                                 int * p_result);
  
void alloc_new_file_block(int nro_physique);
void alloc_new_data_block(int nro_physique);
void desalloc_new_file_block(int nro_physique);
void desalloc_new_data_block(int nro_physique);

error_t update_description_block();

error_t create_file(int   type, 
                    int   permissions, 
                    int * p_nb_physical_block_file);
error_t open_file_from_physical_block_number(int physical_block_number,
                                             int b_reading_authorized, 
                                             int b_writing_authorized, 
                                             int b_writing_at_the_end,
                                             fd * p_fd);
error_t read_file(  fd   file_descr, 
                  void * buffer, 
                   int   nb_bytes_to_read, 
                   int * p_nb_bytes_read);
error_t write_file(  fd   file_descr, 
                   void * buffer, 
                    int   nb_bytes_to_write);

int number_of_data_blocks_for_a_given_file(struct s_fs_inode inode);

int find_nro_current_block(fd file_descr);  
int find_nro_current_byte_in_current_block(fd file_descr);
int calc_number_of_remaining_bytes_in_current_block(fd file_descr);

int min(int i1, int i2);

error_t seek_file(fd  file_descr, 
                  int position);
void close_file(fd file_descr);
error_t stat_file(            fd   file_descr, 
                  struct fs_stat * s_stat);

error_t search_name_in_dir(char * name, 
                            int   dir_physical_block_number, 
                            int * p_physical_block_number);
error_t add_directory_entry(char * file_name, 
                             int   dir_physical_block_number, 
                             int   physical_block_number);
error_t remove_directory_entry(     char * name,  
                               filenum_t   dir_physical_block_number);
error_t readdir_entry(       fd   file_descr,  
                      dir_entry * p_de, 
                            int * p_b_end_of_dir);

error_t truncate_file(fd file_descr, int new_size);
error_t decrease_number_of_links_and_suppress_file_if_necessary(
                                filenum_t physical_block_number);

error_t return_head_and_tail_of_path(char *  path, 
                                     char ** p_head, 
                                     char ** p_tail);

error_t path_to_physical_block_number(char *  path, 
                                       int *  p_physical_block_number,
                                       int    b_file_not_already_created,
                                      char ** p_last_component
                                      );

error_t readdir_entry(       fd   file_descr, 
                      dir_entry * p_de, 
                            int * p_b_end_of_dir);

error_t open_file(char * name,
                   int   b_reading_authorized, 
                   int   b_writing_authorized, 
                   int   b_writing_at_the_end,
                   int   b_create,
                   int   b_directory,
                    fd * p_file_descr);

int decrypt_permissions(int b_writing_authorized, 
                        int b_reading_authorized);

error_t link_file(char * old_path, 
                  char * new_path);
  
int test_read_write_logiques(int size_max_file, 
                             int nb_of_tests);
